/* This file is part of SLB
   Copyright (C) 2011 Sergey Poznyakoff

   SLB is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   SLB is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with SLB.  If not, see <http://www.gnu.org/licenses/>. */

#include "slb.h"

gl_list_t mib_file_list;

struct slb_server *srv_head, *srv_tail;
size_t srv_count;
char *default_expression;

void
register_server(struct slb_server *srv)
{
	if (srv_tail)
		srv_tail->next = srv;
	else
		srv_head = srv;
	srv_tail = srv;
	srv_count++;
}

struct slb_server *
server_lookup(const char *id)
{
	struct slb_server *srv;

	for (srv = srv_head; srv; srv = srv->next)
		if (strcmp(srv->id, id) == 0)
			break;
	return srv;
}


int
assert_string_arg(grecs_locus_t *locus,
		  enum grecs_callback_command cmd,
		  const grecs_value_t *value)
{
	if (cmd != grecs_callback_set_value) {
		grecs_error(locus, 0, _("Unexpected block statement"));
		return 1;
	}
	if (!value || value->type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("expected scalar value as a tag"));
		return 1;
	}
	return 0;
}


static int
cb_mib_directory(enum grecs_callback_command cmd,
		 grecs_locus_t *locus,
		 void *varptr,
		 grecs_value_t *value,
		 void *cb_data)
{
	if (cmd != grecs_callback_set_value) {
		grecs_error(locus, 0, _("Unexpected block statement"));
		return 1;
	}
	if (!value || value->type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("expected string value"));
		return 1;
	}
	add_mibdir(value->v.string);
	return 0;
}

int
cb_define_expression(enum grecs_callback_command cmd,
		     grecs_locus_t *locus,
		     void *varptr,
		     grecs_value_t *value,
		     void *cb_data)
{
	int rc;
	struct slb_expression *expr;
	
	if (cmd != grecs_callback_set_value) {
		grecs_error(locus, 0, _("Unexpected block statement"));
		return 1;
	}
	if (!value || value->type != GRECS_TYPE_ARRAY || value->v.arg.c != 2) {
		grecs_error(locus, 0, _("expected two arguments"));
		return 1;
	}

	if (value->v.arg.v[0].type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("first argument not a string"));
		return 1;
	}

	if (value->v.arg.v[1].type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("second argument not a string"));
		return 1;
	}

	if (compile_expression(value->v.arg.v[1].v.string, locus, &expr))
		return 1;
	expr->ex_name = value->v.arg.v[0].v.string;
	
	rc = register_expression(expr);
	free(expr);
	return rc;
}

int
cb_server_expression(enum grecs_callback_command cmd,
		     grecs_locus_t *locus,
		     void *varptr,
		     grecs_value_t *value,
		     void *cb_data)
{
	struct slb_expression **pexpr = varptr;

	if (cmd != grecs_callback_set_value) {
		grecs_error(locus, 0, _("Unexpected block statement"));
		return 1;
	}
	if (!value || value->type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("expected single string as argument"));
		return 1;
	}

	return compile_expression(value->v.string, locus, pexpr);
}

int
cb_server_variable(enum grecs_callback_command cmd,
		   grecs_locus_t *locus,
		   void *varptr,
		   grecs_value_t *value,
		   void *cb_data)
{
	struct symtab *vt = *(struct symtab **)varptr;
	struct slb_varinstance *inst;
	int install;
	
	if (cmd != grecs_callback_set_value) {
		grecs_error(locus, 0, _("Unexpected block statement"));
		return 1;
	}
	if (!value || value->type != GRECS_TYPE_ARRAY || value->v.arg.c != 2) {
		grecs_error(locus, 0, _("expected two arguments"));
		return 1;
	}

	if (value->v.arg.v[0].type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("first argument not a string"));
		return 1;
	}

	if (value->v.arg.v[1].type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("second argument not a string"));
		return 1;
	}

	install = 1;
	inst = varinst_lookupz(vt, value->v.arg.v[0].v.string, &install);
	if (!install)
		grecs_error(locus, 0,
			     _("redefinition of %s"), inst->vi_sym.name);
	
	inst->vi_oidlen = MAX_OID_LEN;
	if (!read_objid(value->v.arg.v[1].v.string,
			inst->vi_oid, &inst->vi_oidlen)) {
		grecs_error(locus, 0, _("cannot parse oid"));
		return 1;
	}

	inst->vi_mib = value->v.arg.v[1].v.string;

	return 0;
}

int
cb_server_constant(enum grecs_callback_command cmd,
		   grecs_locus_t *locus,
		   void *varptr,
		   grecs_value_t *value,
		   void *cb_data)
{
	struct symtab *vt = *(struct symtab **)varptr;
	struct slb_varinstance *inst;
	int install;
	char *p;
	
	if (cmd != grecs_callback_set_value) {
		grecs_error(locus, 0, _("Unexpected block statement"));
		return 1;
	}
	if (!value || value->type != GRECS_TYPE_ARRAY || value->v.arg.c != 2) {
		grecs_error(locus, 0, _("expected two arguments"));
		return 1;
	}

	if (value->v.arg.v[0].type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("first argument not a string"));
		return 1;
	}

	if (value->v.arg.v[1].type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("second argument not a string"));
		return 1;
	}

	install = 1;
	inst = varinst_lookupz(vt, value->v.arg.v[0].v.string, &install);
	if (!install)
		grecs_error(locus, 0,
			     _("redefinition of %s"), inst->vi_sym.name);
	
	inst->vi_value = strtod(value->v.arg.v[1].v.string, &p);
	if (*p) {
		grecs_error(locus, 0, _("not a valid floating point number"));
		return 1;
	}
	
	inst->vi_flags = SLV_VIF_CONST | SLV_VIF_SET;
	return 0;
}

int
cb_server_assert(enum grecs_callback_command cmd,
		 grecs_locus_t *locus,
		 void *varptr,
		 grecs_value_t *value,
		 void *cb_data)
{
	struct symtab *at = *(struct symtab **)varptr;
	struct slb_assertion *ap;
	int install;
	int flags;
	oid oid[MAX_OID_LEN];
	size_t oidlen;
	size_t len;
	
	if (cmd != grecs_callback_set_value) {
		grecs_error(locus, 0, _("Unexpected block statement"));
		return 1;
	}
	if (!value || value->type != GRECS_TYPE_ARRAY || value->v.arg.c != 3) {
		grecs_error(locus, 0, _("expected two arguments"));
		return 1;
	}

	if (value->v.arg.v[0].type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("first argument not a string"));
		return 1;
	}

	if (value->v.arg.v[1].type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("second argument not a string"));
		return 1;
	}

	if (value->v.arg.v[2].type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("third argument not a string"));
		return 1;
	}

	if (strcmp(value->v.arg.v[1].v.string, "eq") == 0)
		flags = 0;
	else if (strcmp(value->v.arg.v[1].v.string, "ne") == 0)
		flags = SLB_ASSERT_NE;
	else {
		grecs_error(locus, 0, _("invalid operation: %s"),
			    value->v.arg.v[1].v.string);
		return 1;
	}

	oidlen = MAX_OID_LEN;
	if (!read_objid(value->v.arg.v[0].v.string, oid, &oidlen)) {
		grecs_error(locus, 0, _("cannot parse oid"));
		return 1;
	}
	
	install = 1;
	ap = assertion_lookup(at, oid, oidlen, &install);
	if (!install)
		grecs_error(locus, 0,
			    _("redefinition of %s"),
			    value->v.arg.v[0].v.string);

	len = strlen(value->v.arg.v[0].v.string) + 1 +
		strlen(value->v.arg.v[1].v.string) + 1 +
		strlen(value->v.arg.v[2].v.string) + 1;
	ap->text = xmalloc(len);
	strcpy(ap->text, value->v.arg.v[0].v.string);
	strcat(ap->text, " ");
	strcat(ap->text, value->v.arg.v[1].v.string);
	strcat(ap->text, " ");
	strcat(ap->text, value->v.arg.v[2].v.string);

	ap->value = xstrdup(value->v.arg.v[2].v.string);
	ap->flags = flags;
	ap->locus = *locus;
	return 0;
}

int
cb_server_macro(enum grecs_callback_command cmd,
		grecs_locus_t *locus,
		void *varptr,
		grecs_value_t *value,
		void *cb_data)
{
	struct symtab *mt = *(struct symtab **)varptr;
	struct slb_macro *mac;
	int install;
	
	if (cmd != grecs_callback_set_value) {
		grecs_error(locus, 0, _("Unexpected block statement"));
		return 1;
	}
	if (!value || value->type != GRECS_TYPE_ARRAY || value->v.arg.c != 2) {
		grecs_error(locus, 0, _("expected two arguments"));
		return 1;
	}

	if (value->v.arg.v[0].type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("first argument not a string"));
		return 1;
	}

	if (value->v.arg.v[1].type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("second argument not a string"));
		return 1;
	}

	install = 1;
	mac = macro_lookupz(mt, value->v.arg.v[0].v.string, &install);
	if (!install)
		grecs_error(locus, 0,
			    _("redefinition of macro %s"), mac->mac_sym.name);
	
	mac->mac_exp = value->v.arg.v[1].v.string;

	return 0;
}

int
cb_server_enable(enum grecs_callback_command cmd,
		 grecs_locus_t *locus,
		 void *varptr,
		 grecs_value_t *value,
		 void *cb_data)
{
	int *flags = varptr;
	int enable;
	
	if (cmd != grecs_callback_set_value) {
		grecs_error(locus, 0, _("Unexpected block statement"));
		return 1;
	}
	if (!value || value->type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("expected single string as argument"));
		return 1;
	}
	grecs_string_convert(&enable, grecs_type_bool, value->v.string,
			     locus);
	if (!enable)
		*flags |= SLB_SRV_DISABLED;
	return 0;
}

static struct grecs_keyword server_kw[] = {
	{ "enable", NULL, N_("Enable or disable monitoring of this server"),
	  grecs_type_string, NULL, offsetof(struct slb_server, flags),
	  cb_server_enable }, 
	{ "host", NULL, N_("Hostname or IP address"),
	  grecs_type_string, NULL, offsetof(struct slb_server, peer), NULL },
	{ "timeout", NULL, N_("Session timeout"),
	  grecs_type_time, NULL, offsetof(struct slb_server, timeout), NULL },
	{ "retries", NULL, N_("Number of retries before timeout"),
	  grecs_type_uint, NULL, offsetof(struct slb_server, retries), NULL },
	{ "community", NULL, N_("Community string"),
	  grecs_type_string, NULL, offsetof(struct slb_server, community),
	  NULL },
	{ "expression", NULL, N_("Evaluate expression"),
	  grecs_type_string, NULL, offsetof(struct slb_server, expr),
	  cb_server_expression },
	{ "variable", N_("<name: string> <oid: string>"),
	  N_("Define SNMP variable"),
	  grecs_type_string, NULL, offsetof(struct slb_server, varinst),
	  cb_server_variable },
	{ "constant", N_("<name: string> <value: string>"),
	  N_("Define a constant"),
	  grecs_type_string, NULL, offsetof(struct slb_server, varinst),
	  cb_server_constant },
	{ "macro", N_("<name: string> <expansion: string>"),
	  N_("Define a macro"),
	  grecs_type_string, NULL, offsetof(struct slb_server, macros),
	  cb_server_macro },
	{ "assert", N_("<oid: string> {eq|ne} <value: string>"),
	  N_("Ensure that <oid> has the given <value>"),
	  grecs_type_string, NULL, offsetof(struct slb_server, assertions),
	  cb_server_assert },
	{ NULL }
};

static struct slb_server *
slb_server_new(const char *id)
{
	struct slb_server *srv = xzalloc(sizeof(*srv));
	srv->id = xstrdup(id);
	srv->varinst = symtab_create_default(sizeof(struct slb_varinstance));
	srv->macros = symtab_create_default(sizeof(struct slb_macro));
	srv->assertions = assertion_symtab_create();
	return srv;
}

static int
cb_server(enum grecs_callback_command cmd,
	  grecs_locus_t *locus,
	  void *varptr,
	  grecs_value_t *value,
	  void *cb_data)
{
	struct slb_server *srv;
	void **pdata = cb_data;

	switch (cmd) {
	case grecs_callback_section_begin:
		if (!value || value->type != GRECS_TYPE_STRING) {
			grecs_error(locus, 0, _("tag must be a string"));
			return 0;
		}
		srv = slb_server_new(value->v.string);
		srv->locus = *locus;
		*pdata = srv;
		break;

	case grecs_callback_section_end:
		srv = *pdata;
		if (!srv->peer) {
			grecs_error(locus, 0,
				    _("server name not defined"));
			return 1;
		}
		register_server(srv);
		*pdata = NULL;
		break;

	case grecs_callback_set_value:
		grecs_error(locus, 0, _("invalid use of block statement"));
	}
	return 0;
}


static int
cb_syslog_facility(enum grecs_callback_command cmd,
		   grecs_locus_t *locus,
		   void *varptr,
		   grecs_value_t *value,
		   void *cb_data)
{
	if (cmd != grecs_callback_set_value) {
		grecs_error(locus, 0, _("Unexpected block statement"));
		return 1;
	}
	if (!value || value->type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("expected string argument"));
		return 1;
	}

	if (string_to_syslog_facility(value->v.string, varptr))
		grecs_error(locus, 0, _("Unknown syslog facility `%s'"),
			    value->v.string);
	return 0;
}

static struct grecs_keyword syslog_kw[] = {
	{ "facility",
	  N_("name"),
	  N_("Set syslog facility. Arg is one of the following: user, daemon, "
	     "auth, authpriv, mail, cron, local0 through local7 (case-insensitive), "
	     "or a facility number."),
	  grecs_type_string, &syslog_facility, 0, cb_syslog_facility },
	{ "tag", N_("string"), N_("Tag syslog messages with this string"),
	  grecs_type_string, &syslog_tag },
	{ "print-priority", N_("arg"),
	  N_("Prefix each message with its priority"),
	  grecs_type_bool, &syslog_include_prio },
	{ NULL },
};


static struct grecs_keyword slb_kw[] = {
	{ "standalone", NULL, N_("Enable standalone mode"),
	  grecs_type_bool, &standalone_mode },
	{ "foreground", NULL, N_("Start in foreground even in daemon mode"),
	  grecs_type_bool, &foreground },
	{ "pidfile", N_("file"), N_("Set pid file name"),
	  grecs_type_string, &pidfile },
#if 0	
	/* These are reserved for future use.  The '' quotes are used to
	   pacify make check-docs in /dir */
	{ ''user'', N_("name"), N_("Run with UID and GID of this user"),
	  grecs_type_string, NULL, 0, cb_user },
	{ ''group'', NULL, N_("Retain these supplementary groups"),
	  grecs_type_string|GRECS_LIST, NULL, 0, cb_supp_groups },
#endif

	{ "wakeup", NULL, N_("Wake up each <arg> seconds"),
	  grecs_type_time, &wakeup_interval },
	
	{ "mib-directory", N_("name"), N_("Search for MIBs in this directory"),
	  grecs_type_string, NULL, 0, cb_mib_directory },
	{ "add-mib", N_("file"), N_("Read MIBs from file"),
	  grecs_type_string|GRECS_LIST, &mib_file_list },

	{ "expression",
	  N_("<name: string> <expr: text>"),
	  N_("define a named expression"),
	  grecs_type_string, NULL, 0, cb_define_expression },

	{ "default-expression", NULL,
	  N_("Set the name of the default expression"),
	  grecs_type_string, &default_expression },
	  
	{ "server", N_("id: string"), N_("Define a server"),
	  grecs_type_section, NULL, 0, cb_server, NULL, server_kw },

	{ "syslog", NULL, N_("Configure syslog logging"),
	  grecs_type_section, NULL, 0, NULL, NULL, syslog_kw },

	{ "tail", NULL,
	  N_("Output at most <arg> records from the sorted list"),
	  grecs_type_size, &output_tail_count },
	{ "head", NULL,
	  N_("Output at most <arg> records from the sorted list"),
	  grecs_type_size, &output_head_count },
	{ "output-file", NULL,
	  N_("Name of the output file or command"),
	  grecs_type_string, &output_file },
	{ "output-format", NULL,
	  N_("format string for output records"),
	  grecs_type_string, &output_format },
	{ "begin-output-message", NULL,
	  N_("string emitted before formatting output records"),
	  grecs_type_string, &begin_output_message },
	{ "end-output-message", NULL,
	  N_("string emitted after formatting output records"),
	  grecs_type_string, &end_output_message },
	{ "suppress-output", NULL,
	  N_("suppress first <arg> outputs"),
	  grecs_type_size, &output_suppress_count },
	{ NULL }
};

void
config_help()
{
	static char docstring[] =
		N_("Configuration file structure for SLB.\n"
		   "For more information, use `info slb configuration'.");
	grecs_format_docstring (stdout, docstring, 0);
	grecs_format_statement_array (stdout, slb_kw, 1, 0);
}

void
config_init()
{
	grecs_set_keywords(slb_kw);
	grecs_include_path_setup(DEFAULT_VERSION_INCLUDE_DIR,
				 DEFAULT_INCLUDE_DIR, NULL);
	grecs_preprocessor = DEFAULT_PREPROCESSOR;
	grecs_log_to_stderr = true;
}


struct config_finish_hook_entry
{
	struct config_finish_hook_entry *next;
	config_finish_hook_t fun;
	void *data;
};

static struct config_finish_hook_entry *cfh_head, *cfh_tail;
	
void
add_config_finish_hook(config_finish_hook_t fun, void *data)
{
	struct config_finish_hook_entry *ent = xmalloc(sizeof(*ent));
	ent->next = NULL;
	ent->fun = fun;
	ent->data = data;
	if (cfh_tail)
		cfh_tail->next = ent;
	else
		cfh_head = ent;
	cfh_tail = ent;
}

static int
run_config_finish_hooks()
{
	struct config_finish_hook_entry *p;
	int rc = 0;
	
	for (p = cfh_head; p; p = p->next)
		rc |= p->fun(p->data);
	return rc;
}

void
config_finish()
{
	struct slb_server *srv;
	gl_list_iterator_t itr;
	struct slb_expression *defexpr = NULL;
		
	init_mib();

	/* Load requested MIBs */
	if (mib_file_list) {
		const void *p;
		
		itr = gl_list_iterator(mib_file_list);
	
		while (gl_list_iterator_next(&itr, &p, NULL)) {
			read_mib((const char*)p);
		}
		gl_list_iterator_free(&itr);
	}

	collect_output_refs();

	if (default_expression) {
		defexpr = expression_lookupz(default_expression);
		if (!defexpr) {
			grecs_error(NULL, 0,
				    _("default-expression refers to "
				      "undefined expression"));
			exit(EX_CONFIG);
		}
	}

	for (srv = srv_head; srv; srv = srv->next) {
		if (!srv->expr) {
			if (defexpr)
				srv->expr = copy_expression(defexpr, NULL);
			else {
				grecs_error(&srv->locus, 0,
					    _("server expression not defined"));
				continue;
			}
		}
		if (fixup_expression(srv->expr, srv->varinst))
			srv->flags |= SLB_SRV_DISABLED;
		else {
			report_unbound_output_refs(srv);
			srv->oidtab = create_oidtab(srv->varinst);
		}
	}

	if (grecs_error_count || run_config_finish_hooks())
		exit(EX_CONFIG);
}	
